@@ -9,7 +9,8 @@ use solana_sdk::{
     account_utils::StateMut,
     bpf_loader_upgradeable::{self, UpgradeableLoaderState},
     feature_set::{
-        instructions_sysvar_enabled, neon_evm_compute_budget, updated_verify_policy, FeatureSet,
+        demote_program_write_locks, fix_write_privs, instructions_sysvar_enabled,
+        neon_evm_compute_budget, updated_verify_policy, FeatureSet,
     },
     ic_logger_msg, ic_msg,
     instruction::{CompiledInstruction, Instruction, InstructionError},
@@ -300,6 +301,7 @@ impl<'a> ThisInvokeContext<'a> {
             instruction,
             executable_accounts,
             accounts,
+            feature_set.is_active(&demote_program_write_locks::id()),
         );
         let mut invoke_context = Self {
             invoke_stack: Vec::with_capacity(bpf_compute_budget.max_invoke_depth),
@@ -611,6 +613,7 @@ impl MessageProcessor {
         instruction: &'a CompiledInstruction,
         executable_accounts: &'a [(Pubkey, Rc<RefCell<AccountSharedData>>)],
         accounts: &'a [(Pubkey, Rc<RefCell<AccountSharedData>>)],
+        demote_program_write_locks: bool,
     ) -> Vec<(bool, bool, &'a Pubkey, &'a RefCell<AccountSharedData>)> {
         executable_accounts
             .iter()
@@ -619,7 +622,7 @@ impl MessageProcessor {
                 let index = *index as usize;
                 (
                     message.is_signer(index),
-                    message.is_writable(index),
+                    message.is_writable(index, demote_program_write_locks),
                     &accounts[index].0,
                     &accounts[index].1 as &RefCell<AccountSharedData>,
                 )
@@ -759,38 +762,64 @@ impl MessageProcessor {
         ) = {
             let invoke_context = invoke_context.borrow();
 
-            // Translate and verify caller's data
-            let keyed_accounts = invoke_context.get_keyed_accounts()?;
-            let keyed_accounts = keyed_account_indices
+            let caller_keyed_accounts = invoke_context.get_keyed_accounts()?;
+            let callee_keyed_accounts = keyed_account_indices
                 .iter()
-                .map(|index| keyed_account_at_index(keyed_accounts, *index))
+                .map(|index| keyed_account_at_index(caller_keyed_accounts, *index))
                 .collect::<Result<Vec<&KeyedAccount>, InstructionError>>()?;
-            let (message, callee_program_id, _) =
-                Self::create_message(&instruction, &keyed_accounts, signers, &invoke_context)?;
-            let keyed_accounts = invoke_context.get_keyed_accounts()?;
-            let mut caller_write_privileges = keyed_account_indices
-                .iter()
-                .map(|index| keyed_accounts[*index].is_writable())
-                .collect::<Vec<bool>>();
-            caller_write_privileges.insert(0, false);
-            let mut accounts = vec![];
-            let mut keyed_account_indices_reordered = vec![];
-            let keyed_accounts = invoke_context.get_keyed_accounts()?;
-            'root: for account_key in message.account_keys.iter() {
-                for keyed_account_index in keyed_account_indices {
-                    let keyed_account = &keyed_accounts[*keyed_account_index];
-                    if account_key == keyed_account.unsigned_key() {
-                        accounts.push((*account_key, Rc::new(keyed_account.account.clone())));
-                        keyed_account_indices_reordered.push(*keyed_account_index);
-                        continue 'root;
+            let (message, callee_program_id, _) = Self::create_message(
+                &instruction,
+                &callee_keyed_accounts,
+                signers,
+                &invoke_context,
+            )?;
+            let mut keyed_account_indices_reordered =
+                Vec::with_capacity(message.account_keys.len());
+            let mut accounts = Vec::with_capacity(message.account_keys.len());
+            let mut caller_write_privileges = Vec::with_capacity(message.account_keys.len());
+
+            // Translate and verify caller's data
+            if invoke_context.is_feature_active(&fix_write_privs::id()) {
+                'root: for account_key in message.account_keys.iter() {
+                    for keyed_account_index in keyed_account_indices {
+                        let keyed_account = &caller_keyed_accounts[*keyed_account_index];
+                        if account_key == keyed_account.unsigned_key() {
+                            accounts.push((*account_key, Rc::new(keyed_account.account.clone())));
+                            caller_write_privileges.push(keyed_account.is_writable());
+                            keyed_account_indices_reordered.push(*keyed_account_index);
+                            continue 'root;
+                        }
                     }
+                    ic_msg!(
+                        invoke_context,
+                        "Instruction references an unknown account {}",
+                        account_key
+                    );
+                    return Err(InstructionError::MissingAccount);
+                }
+            } else {
+                let keyed_accounts = invoke_context.get_keyed_accounts()?;
+                for index in keyed_account_indices.iter() {
+                    caller_write_privileges.push(keyed_accounts[*index].is_writable());
+                }
+                caller_write_privileges.insert(0, false);
+                let keyed_accounts = invoke_context.get_keyed_accounts()?;
+                'root2: for account_key in message.account_keys.iter() {
+                    for keyed_account_index in keyed_account_indices {
+                        let keyed_account = &keyed_accounts[*keyed_account_index];
+                        if account_key == keyed_account.unsigned_key() {
+                            accounts.push((*account_key, Rc::new(keyed_account.account.clone())));
+                            keyed_account_indices_reordered.push(*keyed_account_index);
+                            continue 'root2;
+                        }
+                    }
+                    ic_msg!(
+                        invoke_context,
+                        "Instruction references an unknown account {}",
+                        account_key
+                    );
+                    return Err(InstructionError::MissingAccount);
                 }
-                ic_msg!(
-                    invoke_context,
-                    "Instruction references an unknown account {}",
-                    account_key
-                );
-                return Err(InstructionError::MissingAccount);
             }
 
             // Process instruction
@@ -864,6 +893,8 @@ impl MessageProcessor {
 
         {
             let invoke_context = invoke_context.borrow();
+            let demote_program_write_locks =
+                invoke_context.is_feature_active(&demote_program_write_locks::id());
             let keyed_accounts = invoke_context.get_keyed_accounts()?;
             for (src_keyed_account_index, ((_key, account), dst_keyed_account_index)) in accounts
                 .iter()
@@ -872,7 +903,9 @@ impl MessageProcessor {
             {
                 let dst_keyed_account = &keyed_accounts[dst_keyed_account_index];
                 let src_keyed_account = account.borrow();
-                if message.is_writable(src_keyed_account_index) && !src_keyed_account.executable() {
+                if message.is_writable(src_keyed_account_index, demote_program_write_locks)
+                    && !src_keyed_account.executable()
+                {
                     if dst_keyed_account.data_len()? != src_keyed_account.data().len()
                         && dst_keyed_account.data_len()? != 0
                     {
@@ -916,8 +949,15 @@ impl MessageProcessor {
             invoke_context.verify_and_update(instruction, accounts, caller_write_privileges)?;
 
             // Construct keyed accounts
-            let keyed_accounts =
-                Self::create_keyed_accounts(message, instruction, executable_accounts, accounts);
+            let demote_program_write_locks =
+                invoke_context.is_feature_active(&demote_program_write_locks::id());
+            let keyed_accounts = Self::create_keyed_accounts(
+                message,
+                instruction,
+                executable_accounts,
+                accounts,
+                demote_program_write_locks,
+            );
 
             // Invoke callee
             invoke_context.push(program_id, &keyed_accounts)?;
@@ -935,7 +975,7 @@ impl MessageProcessor {
             if result.is_ok() {
                 // Verify the called program has not misbehaved
                 let write_privileges: Vec<bool> = (0..message.account_keys.len())
-                    .map(|i| message.is_writable(i))
+                    .map(|i| message.is_writable(i, demote_program_write_locks))
                     .collect();
                 result = invoke_context.verify_and_update(instruction, accounts, &write_privileges);
             }
@@ -984,6 +1024,7 @@ impl MessageProcessor {
     }
 
     /// Verify the results of an instruction
+    #[allow(clippy::too_many_arguments)]
     pub fn verify(
         message: &Message,
         instruction: &CompiledInstruction,
@@ -994,6 +1035,7 @@ impl MessageProcessor {
         timings: &mut ExecuteDetailsTimings,
         logger: Rc<RefCell<dyn Logger>>,
         updated_verify_policy: bool,
+        demote_program_write_locks: bool,
     ) -> Result<(), InstructionError> {
         // Verify all executable accounts have zero outstanding refs
         Self::verify_account_references(executable_accounts)?;
@@ -1014,7 +1056,7 @@ impl MessageProcessor {
                 pre_accounts[unique_index]
                     .verify(
                         program_id,
-                        message.is_writable(account_index),
+                        message.is_writable(account_index, demote_program_write_locks),
                         rent,
                         &account,
                         timings,
@@ -1183,6 +1225,7 @@ impl MessageProcessor {
             timings,
             invoke_context.get_logger(),
             invoke_context.is_feature_active(&updated_verify_policy::id()),
+            invoke_context.is_feature_active(&demote_program_write_locks::id()),
         )?;
 
         timings.accumulate(&invoke_context.timings);
@@ -1338,7 +1381,7 @@ mod tests {
                 ))),
             ));
             let write_privileges: Vec<bool> = (0..message.account_keys.len())
-                .map(|i| message.is_writable(i))
+                .map(|i| message.is_writable(i, /*demote_program_write_locks=*/ true))
                 .collect();
             invoke_context
                 .verify_and_update(&message.instructions[0], &these_accounts, &write_privileges)
@@ -2163,6 +2206,7 @@ mod tests {
             NoopFail,
             ModifyOwned,
             ModifyNotOwned,
+            ModifyReadonly,
         }
 
         fn mock_process_instruction(
@@ -2187,6 +2231,9 @@ mod tests {
                     MockInstruction::ModifyNotOwned => {
                         keyed_accounts[1].try_account_ref_mut()?.data_as_mut_slice()[0] = 1
                     }
+                    MockInstruction::ModifyReadonly => {
+                        keyed_accounts[2].try_account_ref_mut()?.data_as_mut_slice()[0] = 1
+                    }
                 }
             } else {
                 return Err(InstructionError::InvalidInstructionData);
@@ -2206,6 +2253,7 @@ mod tests {
 
         let owned_account = AccountSharedData::new(42, 1, &callee_program_id);
         let not_owned_account = AccountSharedData::new(84, 1, &solana_sdk::pubkey::new_rand());
+        let readonly_account = AccountSharedData::new(168, 1, &caller_program_id);
 
         #[allow(unused_mut)]
         let accounts = vec![
@@ -2217,30 +2265,38 @@ mod tests {
                 solana_sdk::pubkey::new_rand(),
                 Rc::new(RefCell::new(not_owned_account)),
             ),
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(readonly_account)),
+            ),
             (callee_program_id, Rc::new(RefCell::new(program_account))),
         ];
 
-        let compiled_instruction = CompiledInstruction::new(2, &(), vec![0, 1, 2]);
         let programs: Vec<(_, ProcessInstructionWithContext)> =
             vec![(callee_program_id, mock_process_instruction)];
         let metas = vec![
             AccountMeta::new(accounts[0].0, false),
             AccountMeta::new(accounts[1].0, false),
+            AccountMeta::new_readonly(accounts[2].0, false),
         ];
 
-        let instruction = Instruction::new_with_bincode(
+        let caller_instruction = CompiledInstruction::new(2, &(), vec![0, 1, 2, 3]);
+        let callee_instruction = Instruction::new_with_bincode(
             callee_program_id,
             &MockInstruction::NoopSuccess,
             metas.clone(),
         );
-        let message = Message::new(&[instruction], None);
+        let message = Message::new(&[callee_instruction], None);
+
+        let feature_set = FeatureSet::all_enabled();
+        let demote_program_write_locks = feature_set.is_active(&demote_program_write_locks::id());
 
         let ancestors = Ancestors::default();
         let mut invoke_context = ThisInvokeContext::new(
             &caller_program_id,
             Rent::default(),
             &message,
-            &compiled_instruction,
+            &caller_instruction,
             &executable_accounts,
             &accounts,
             programs.as_slice(),
@@ -2248,7 +2304,7 @@ mod tests {
             BpfComputeBudget::default(),
             Rc::new(RefCell::new(Executors::default())),
             None,
-            Arc::new(FeatureSet::all_enabled()),
+            Arc::new(feature_set),
             Arc::new(Accounts::default()),
             &ancestors,
         );
@@ -2258,7 +2314,7 @@ mod tests {
             .account_keys
             .iter()
             .enumerate()
-            .map(|(i, _)| message.is_writable(i))
+            .map(|(i, _)| message.is_writable(i, demote_program_write_locks))
             .collect::<Vec<bool>>();
         accounts[0].1.borrow_mut().data_as_mut_slice()[0] = 1;
         assert_eq!(
@@ -2273,6 +2329,20 @@ mod tests {
         );
         accounts[0].1.borrow_mut().data_as_mut_slice()[0] = 0;
 
+        // readonly account modified by the invoker
+        accounts[2].1.borrow_mut().data_as_mut_slice()[0] = 1;
+        assert_eq!(
+            MessageProcessor::process_cross_program_instruction(
+                &message,
+                &executable_accounts,
+                &accounts,
+                &caller_write_privileges,
+                &mut invoke_context,
+            ),
+            Err(InstructionError::ReadonlyDataModified)
+        );
+        accounts[2].1.borrow_mut().data_as_mut_slice()[0] = 0;
+
         let cases = vec![
             (MockInstruction::NoopSuccess, Ok(())),
             (
@@ -2287,16 +2357,16 @@ mod tests {
         ];
 
         for case in cases {
-            let instruction =
+            let callee_instruction =
                 Instruction::new_with_bincode(callee_program_id, &case.0, metas.clone());
-            let message = Message::new(&[instruction], None);
+            let message = Message::new(&[callee_instruction], None);
 
             let ancestors = Ancestors::default();
             let mut invoke_context = ThisInvokeContext::new(
                 &caller_program_id,
                 Rent::default(),
                 &message,
-                &compiled_instruction,
+                &caller_instruction,
                 &executable_accounts,
                 &accounts,
                 programs.as_slice(),
@@ -2313,7 +2383,7 @@ mod tests {
                 .account_keys
                 .iter()
                 .enumerate()
-                .map(|(i, _)| message.is_writable(i))
+                .map(|(i, _)| message.is_writable(i, demote_program_write_locks))
                 .collect::<Vec<bool>>();
             assert_eq!(
                 MessageProcessor::process_cross_program_instruction(
@@ -2353,4 +2423,189 @@ mod tests {
 
         assert!(!format!("{:?}", message_processor).is_empty());
     }
+
+    #[test]
+    fn test_native_invoke() {
+        #[derive(Debug, Serialize, Deserialize)]
+        enum MockInstruction {
+            NoopSuccess,
+            NoopFail,
+            ModifyOwned,
+            ModifyNotOwned,
+            ModifyReadonly,
+        }
+
+        fn mock_process_instruction(
+            program_id: &Pubkey,
+            data: &[u8],
+            invoke_context: &mut dyn InvokeContext,
+        ) -> Result<(), InstructionError> {
+            let keyed_accounts = invoke_context.get_keyed_accounts()?;
+            assert_eq!(*program_id, keyed_accounts[0].owner()?);
+            assert_ne!(
+                keyed_accounts[1].owner()?,
+                *keyed_accounts[0].unsigned_key()
+            );
+
+            if let Ok(instruction) = bincode::deserialize(data) {
+                match instruction {
+                    MockInstruction::NoopSuccess => (),
+                    MockInstruction::NoopFail => return Err(InstructionError::GenericError),
+                    MockInstruction::ModifyOwned => {
+                        keyed_accounts[0].try_account_ref_mut()?.data_as_mut_slice()[0] = 1
+                    }
+                    MockInstruction::ModifyNotOwned => {
+                        keyed_accounts[1].try_account_ref_mut()?.data_as_mut_slice()[0] = 1
+                    }
+                    MockInstruction::ModifyReadonly => {
+                        keyed_accounts[2].try_account_ref_mut()?.data_as_mut_slice()[0] = 1
+                    }
+                }
+            } else {
+                return Err(InstructionError::InvalidInstructionData);
+            }
+            Ok(())
+        }
+
+        let caller_program_id = solana_sdk::pubkey::new_rand();
+        let callee_program_id = solana_sdk::pubkey::new_rand();
+
+        let mut program_account = AccountSharedData::new(1, 0, &native_loader::id());
+        program_account.set_executable(true);
+        let executable_accounts = vec![(
+            callee_program_id,
+            Rc::new(RefCell::new(program_account.clone())),
+        )];
+
+        let owned_account = AccountSharedData::new(42, 1, &callee_program_id);
+        let not_owned_account = AccountSharedData::new(84, 1, &solana_sdk::pubkey::new_rand());
+        let readonly_account = AccountSharedData::new(168, 1, &caller_program_id);
+
+        #[allow(unused_mut)]
+        let accounts = vec![
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(owned_account)),
+            ),
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(not_owned_account)),
+            ),
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(readonly_account)),
+            ),
+            (callee_program_id, Rc::new(RefCell::new(program_account))),
+        ];
+        let programs: Vec<(_, ProcessInstructionWithContext)> =
+            vec![(callee_program_id, mock_process_instruction)];
+        let metas = vec![
+            AccountMeta::new(accounts[0].0, false),
+            AccountMeta::new(accounts[1].0, false),
+            AccountMeta::new_readonly(accounts[2].0, false),
+        ];
+
+        let caller_instruction = CompiledInstruction::new(2, &(), vec![0, 1, 2, 3]);
+        let callee_instruction = Instruction::new_with_bincode(
+            callee_program_id,
+            &MockInstruction::NoopSuccess,
+            metas.clone(),
+        );
+        let message = Message::new(&[callee_instruction.clone()], None);
+
+        let ancestors = Ancestors::default();
+        let mut invoke_context = ThisInvokeContext::new(
+            &caller_program_id,
+            Rent::default(),
+            &message,
+            &caller_instruction,
+            &executable_accounts,
+            &accounts,
+            programs.as_slice(),
+            None,
+            BpfComputeBudget::default(),
+            Rc::new(RefCell::new(Executors::default())),
+            None,
+            Arc::new(FeatureSet::all_enabled()),
+            Arc::new(Accounts::default()),
+            &ancestors,
+        );
+
+        // not owned account modified by the invoker
+        accounts[0].1.borrow_mut().data_as_mut_slice()[0] = 1;
+        assert_eq!(
+            MessageProcessor::native_invoke(
+                &mut invoke_context,
+                callee_instruction.clone(),
+                &[0, 1, 2, 3],
+                &[]
+            ),
+            Err(InstructionError::ExternalAccountDataModified)
+        );
+        accounts[0].1.borrow_mut().data_as_mut_slice()[0] = 0;
+
+        // readonly account modified by the invoker
+        accounts[2].1.borrow_mut().data_as_mut_slice()[0] = 1;
+        assert_eq!(
+            MessageProcessor::native_invoke(
+                &mut invoke_context,
+                callee_instruction,
+                &[0, 1, 2, 3],
+                &[]
+            ),
+            Err(InstructionError::ReadonlyDataModified)
+        );
+        accounts[2].1.borrow_mut().data_as_mut_slice()[0] = 0;
+
+        // Other test cases
+        let cases = vec![
+            (MockInstruction::NoopSuccess, Ok(())),
+            (
+                MockInstruction::NoopFail,
+                Err(InstructionError::GenericError),
+            ),
+            (MockInstruction::ModifyOwned, Ok(())),
+            (
+                MockInstruction::ModifyNotOwned,
+                Err(InstructionError::ExternalAccountDataModified),
+            ),
+            (
+                MockInstruction::ModifyReadonly,
+                Err(InstructionError::ReadonlyDataModified),
+            ),
+        ];
+        for case in cases {
+            let callee_instruction =
+                Instruction::new_with_bincode(callee_program_id, &case.0, metas.clone());
+            let message = Message::new(&[callee_instruction.clone()], None);
+
+            let ancestors = Ancestors::default();
+            let mut invoke_context = ThisInvokeContext::new(
+                &caller_program_id,
+                Rent::default(),
+                &message,
+                &caller_instruction,
+                &executable_accounts,
+                &accounts,
+                programs.as_slice(),
+                None,
+                BpfComputeBudget::default(),
+                Rc::new(RefCell::new(Executors::default())),
+                None,
+                Arc::new(FeatureSet::all_enabled()),
+                Arc::new(Accounts::default()),
+                &ancestors,
+            );
+
+            assert_eq!(
+                MessageProcessor::native_invoke(
+                    &mut invoke_context,
+                    callee_instruction,
+                    &[0, 1, 2, 3],
+                    &[]
+                ),
+                case.1
+            );
+        }
+    }
 }
