@@ -121,6 +121,7 @@ impl Stakes {
         pubkey: &Pubkey,
         account: &AccountSharedData,
         check_vote_init: bool,
+        remove_delegation_on_inactive: bool,
     ) -> Option<ArcVoteAccount> {
         if solana_vote_program::check_id(account.owner()) {
             // unconditionally remove existing at first; there is no dependent calculated state for
@@ -176,7 +177,13 @@ impl Stakes {
                 }
             }
 
-            if account.lamports() == 0 {
+            let remove_delegation = if remove_delegation_on_inactive {
+                delegation.is_none()
+            } else {
+                account.lamports() == 0
+            };
+
+            if remove_delegation {
                 // when account is removed (lamports == 0), remove it from Stakes as well
                 // so that given `pubkey` can be used for any owner in the future, while not
                 // affecting Stakes.
@@ -295,8 +302,8 @@ pub mod tests {
             let ((vote_pubkey, vote_account), (stake_pubkey, mut stake_account)) =
                 create_staked_node_accounts(10);
 
-            stakes.store(&vote_pubkey, &vote_account, true);
-            stakes.store(&stake_pubkey, &stake_account, true);
+            stakes.store(&vote_pubkey, &vote_account, true, true);
+            stakes.store(&stake_pubkey, &stake_account, true, true);
             let stake = stake_state::stake_from(&stake_account).unwrap();
             {
                 let vote_accounts = stakes.vote_accounts();
@@ -308,7 +315,7 @@ pub mod tests {
             }
 
             stake_account.set_lamports(42);
-            stakes.store(&stake_pubkey, &stake_account, true);
+            stakes.store(&stake_pubkey, &stake_account, true, true);
             {
                 let vote_accounts = stakes.vote_accounts();
                 assert!(vote_accounts.get(&vote_pubkey).is_some());
@@ -320,7 +327,7 @@ pub mod tests {
 
             // activate more
             let (_stake_pubkey, mut stake_account) = create_stake_account(42, &vote_pubkey);
-            stakes.store(&stake_pubkey, &stake_account, true);
+            stakes.store(&stake_pubkey, &stake_account, true, true);
             let stake = stake_state::stake_from(&stake_account).unwrap();
             {
                 let vote_accounts = stakes.vote_accounts();
@@ -332,7 +339,7 @@ pub mod tests {
             }
 
             stake_account.set_lamports(0);
-            stakes.store(&stake_pubkey, &stake_account, true);
+            stakes.store(&stake_pubkey, &stake_account, true, true);
             {
                 let vote_accounts = stakes.vote_accounts();
                 assert!(vote_accounts.get(&vote_pubkey).is_some());
@@ -350,14 +357,14 @@ pub mod tests {
         let ((vote_pubkey, vote_account), (stake_pubkey, stake_account)) =
             create_staked_node_accounts(10);
 
-        stakes.store(&vote_pubkey, &vote_account, true);
-        stakes.store(&stake_pubkey, &stake_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
 
         let ((vote11_pubkey, vote11_account), (stake11_pubkey, stake11_account)) =
             create_staked_node_accounts(20);
 
-        stakes.store(&vote11_pubkey, &vote11_account, true);
-        stakes.store(&stake11_pubkey, &stake11_account, true);
+        stakes.store(&vote11_pubkey, &vote11_account, true, true);
+        stakes.store(&stake11_pubkey, &stake11_account, true, true);
 
         let vote11_node_pubkey = VoteState::from(&vote11_account).unwrap().node_pubkey;
 
@@ -374,8 +381,8 @@ pub mod tests {
         let ((vote_pubkey, mut vote_account), (stake_pubkey, stake_account)) =
             create_staked_node_accounts(10);
 
-        stakes.store(&vote_pubkey, &vote_account, true);
-        stakes.store(&stake_pubkey, &stake_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -384,7 +391,7 @@ pub mod tests {
         }
 
         vote_account.set_lamports(0);
-        stakes.store(&vote_pubkey, &vote_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -392,7 +399,7 @@ pub mod tests {
         }
 
         vote_account.set_lamports(1);
-        stakes.store(&vote_pubkey, &vote_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -405,7 +412,7 @@ pub mod tests {
         let mut pushed = vote_account.data().to_vec();
         pushed.push(0);
         vote_account.set_data(pushed);
-        stakes.store(&vote_pubkey, &vote_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -416,7 +423,7 @@ pub mod tests {
         let default_vote_state = VoteState::default();
         let versioned = VoteStateVersions::new_current(default_vote_state);
         VoteState::to(&versioned, &mut vote_account).unwrap();
-        stakes.store(&vote_pubkey, &vote_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -424,7 +431,7 @@ pub mod tests {
         }
 
         vote_account.set_data(cache_data);
-        stakes.store(&vote_pubkey, &vote_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -446,11 +453,11 @@ pub mod tests {
         let ((vote_pubkey2, vote_account2), (_stake_pubkey2, stake_account2)) =
             create_staked_node_accounts(10);
 
-        stakes.store(&vote_pubkey, &vote_account, true);
-        stakes.store(&vote_pubkey2, &vote_account2, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&vote_pubkey2, &vote_account2, true, true);
 
         // delegates to vote_pubkey
-        stakes.store(&stake_pubkey, &stake_account, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
 
         let stake = stake_state::stake_from(&stake_account).unwrap();
 
@@ -466,7 +473,7 @@ pub mod tests {
         }
 
         // delegates to vote_pubkey2
-        stakes.store(&stake_pubkey, &stake_account2, true);
+        stakes.store(&stake_pubkey, &stake_account2, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -491,11 +498,11 @@ pub mod tests {
 
         let (stake_pubkey2, stake_account2) = create_stake_account(10, &vote_pubkey);
 
-        stakes.store(&vote_pubkey, &vote_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
 
         // delegates to vote_pubkey
-        stakes.store(&stake_pubkey, &stake_account, true);
-        stakes.store(&stake_pubkey2, &stake_account2, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
+        stakes.store(&stake_pubkey2, &stake_account2, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -510,8 +517,8 @@ pub mod tests {
         let ((vote_pubkey, vote_account), (stake_pubkey, stake_account)) =
             create_staked_node_accounts(10);
 
-        stakes.store(&vote_pubkey, &vote_account, true);
-        stakes.store(&stake_pubkey, &stake_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
         let stake = stake_state::stake_from(&stake_account).unwrap();
 
         {
@@ -541,8 +548,8 @@ pub mod tests {
         let ((vote_pubkey, vote_account), (stake_pubkey, stake_account)) =
             create_staked_node_accounts(10);
 
-        stakes.store(&vote_pubkey, &vote_account, true);
-        stakes.store(&stake_pubkey, &stake_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
 
         {
             let vote_accounts = stakes.vote_accounts();
@@ -555,6 +562,7 @@ pub mod tests {
             &stake_pubkey,
             &AccountSharedData::new(1, 0, &stake::program::id()),
             true,
+            true,
         );
         {
             let vote_accounts = stakes.vote_accounts();
@@ -584,8 +592,8 @@ pub mod tests {
         let genesis_epoch = 0;
         let ((vote_pubkey, vote_account), (stake_pubkey, stake_account)) =
             create_warming_staked_node_accounts(10, genesis_epoch);
-        stakes.store(&vote_pubkey, &vote_account, true);
-        stakes.store(&stake_pubkey, &stake_account, true);
+        stakes.store(&vote_pubkey, &vote_account, true, true);
+        stakes.store(&stake_pubkey, &stake_account, true, true);
 
         assert_eq!(stakes.vote_balance_and_staked(), 11);
         assert_eq!(stakes.vote_balance_and_warmed_staked(), 1);
